# Copyright (C) 2012 - 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""
Classes and functions for working with geometries in KML files.

The classes in this module represent different types of geometries, such as points,
lines, polygons, and multi-geometries.
These geometries can be used to define the shape and location of features in KML files.

The module also provides functions for handling coordinates and extracting them from XML
elements.

"""

import logging
import re
from typing import Any
from typing import Dict
from typing import Final
from typing import Iterable
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import pygeoif.geometry as geo
from pygeoif.exceptions import DimensionError
from pygeoif.factories import shape
from pygeoif.types import GeoCollectionType
from pygeoif.types import GeoType
from pygeoif.types import LineType
from typing_extensions import Self

from fastkml import config
from fastkml.base import _XMLObject
from fastkml.enums import AltitudeMode
from fastkml.enums import Verbosity
from fastkml.exceptions import GeometryError
from fastkml.exceptions import KMLParseError
from fastkml.exceptions import KMLWriteError
from fastkml.helpers import bool_subelement
from fastkml.helpers import enum_subelement
from fastkml.helpers import subelement_bool_kwarg
from fastkml.helpers import subelement_enum_kwarg
from fastkml.helpers import xml_subelement
from fastkml.helpers import xml_subelement_kwarg
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.registry import RegistryItem
from fastkml.registry import registry
from fastkml.types import Element

__all__ = [
    "AnyGeometryType",
    "Coordinates",
    "GeometryType",
    "LineString",
    "LinearRing",
    "MultiGeometry",
    "MultiGeometryType",
    "Point",
    "Polygon",
    "create_multigeometry",
]

logger = logging.getLogger(__name__)

GeometryType = Union[geo.Polygon, geo.LineString, geo.LinearRing, geo.Point]
MultiGeometryType = Union[
    geo.MultiPoint,
    geo.MultiLineString,
    geo.MultiPolygon,
    geo.GeometryCollection,
]
AnyGeometryType = Union[GeometryType, MultiGeometryType]

MsgMutualExclusive: Final = "Geometry and kml coordinates are mutually exclusive"

xml_attrs = {"ns", "name_spaces", "id", "target_id"}


def handle_invalid_geometry_error(
    *,
    error: Exception,
    element: Element,
    strict: bool,
) -> None:
    """
    Handle an invalid geometry error.

    Args:
    ----
        error (Exception): The exception that occurred.
        element (Element): The XML element that caused the error.
        strict (bool): Flag indicating whether to raise an exception or not.

    Returns:
    -------
        None

    Raises:
    ------
        KMLParseError: If `strict` is True, raise a KMLParseError.

    """
    error_in_xml = config.etree.tostring(
        element,
        encoding="UTF-8",
    ).decode(
        "UTF-8",
    )
    msg = f"Invalid coordinates in '{error_in_xml}' caused by '{error}'"
    logger.error(msg)
    if strict:
        raise KMLParseError(msg) from error


def coordinates_subelement(
    obj: _XMLObject,
    *,
    element: Element,
    attr_name: str,
    node_name: str,  # noqa: ARG001
    precision: Optional[int],
    verbosity: Optional[Verbosity],  # noqa: ARG001
    default: Any,  # noqa: ARG001
) -> None:
    """
    Set the value of an attribute from a subelement with a text node.

    Args:
    ----
        obj (_XMLObject): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the subelement to create.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Any): The default value of the attribute (unused).

    Returns:
    -------
        None

    """
    if getattr(obj, attr_name, None):
        coords = getattr(obj, attr_name)
        if not coords or len(coords[0]) not in (2, 3):
            msg = f"Invalid dimensions in coordinates '{coords}'"
            raise KMLWriteError(msg)
        if precision is None:
            tuples = (",".join(str(c) for c in coord) for coord in coords)
        else:
            tuples = (",".join(f"{c:.{precision}f}" for c in coord) for coord in coords)
        element.text = " ".join(tuples)


def subelement_coordinates_kwarg(
    *,
    element: Element,
    ns: str,  # noqa: ARG001
    name_spaces: Dict[str, str],  # noqa: ARG001
    node_name: str,  # noqa: ARG001
    kwarg: str,
    classes: Tuple[Type[object], ...],  # noqa: ARG001
    strict: bool,
) -> Dict[str, LineType]:
    """
    Extract coordinates from a subelement and returns them as a dictionary.

    Args:
    ----
        element (Element): The XML element containing the coordinates.
        ns (str): The namespace of the XML element.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the XML node containing the coordinates.
        kwarg (str): The name of the keyword argument to store the coordinates.
        classes (Tuple[Type[object], ...]): A tuple of known types for validation.
        strict (bool): A flag indicating whether to raise an error for invalid geometry.

    Returns:
    -------
        Dict[str, LineType]: A dictionary containing the extracted coordinates.

    Raises:
    ------
        ValueError: If the coordinates are not in the expected format.

    """
    try:
        latlons = re.sub(r", +", ",", element.text.strip()).split()
    except AttributeError:
        return {}
    try:
        return {
            kwarg: [  # type: ignore[dict-item]
                tuple(float(c) for c in latlon.split(",")) for latlon in latlons
            ],
        }
    except ValueError as error:
        handle_invalid_geometry_error(
            error=error,
            element=element,
            strict=strict,
        )
        return {}


class Coordinates(_XMLObject):
    """
    Represents a set of coordinates in decimal degrees.

    Attributes
    ----------
        coords (LineType): A list of tuples representing the coordinates.
            Each coord consists of floating point values for
            longitude, latitude, and altitude.
            The altitude component is optional.
            Coordinates are expressed in decimal degrees only.

    """

    _default_nsid = config.KML
    coords: LineType

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        coords: Optional[LineType] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Geometry object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the element.
        name_spaces : dict, optional
            A dictionary of namespace prefixes and URIs.
        coords : LineType, optional
            The coordinates of the geometry.
        **kwargs : dict
            Additional keyword arguments.

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.coords = coords or []

    def __repr__(self) -> str:
        """Create a string (c)representation for Coordinates."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"coords={self.coords!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the geometry has any coordinates.

        Returns
        -------
        bool
            True if the geometry has coordinates, False otherwise.

        """
        return bool(self.coords)

    @classmethod
    def get_tag_name(cls) -> str:
        """Return the tag name."""
        return cls.__name__.lower()


registry.register(
    Coordinates,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(LineType,),  # type: ignore[arg-type]
        attr_name="coords",
        node_name="coordinates",
        get_kwarg=subelement_coordinates_kwarg,
        set_element=coordinates_subelement,
    ),
)


class _Geometry(_BaseObject):
    """
    Baseclass with common methods for all geometry objects.

    Attributes: extrude: boolean --> Specifies whether to connect the feature to
                                     the ground with a line.
                tessellate: boolean -->  Specifies whether to allow the LineString
                                         to follow the terrain.
                altitudeMode: --> Specifies how altitude components in the <coordinates>
                                  element are interpreted.

    """

    altitude_mode: Optional[AltitudeMode]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a _Geometry object.

        Args:
        ----
            ns: Namespace of the object.
            name_spaces: Name spaces of the object.
            id: Id of the object.
            target_id: Target id of the object.
            extrude: Specifies whether to connect the feature to the ground with a line.
            tessellate: Specifies whether to allow the LineString to follow the terrain.
            altitude_mode: Specifies how altitude components in the <coordinates>
                           element are interpreted.
            **kwargs: Additional keyword arguments.

        """
        super().__init__(
            ns=ns,
            id=id,
            name_spaces=name_spaces,
            target_id=target_id,
            **kwargs,
        )
        self.altitude_mode = altitude_mode

    def __repr__(self) -> str:
        """Create a string (c)representation for _Geometry."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"**{self._get_splat()!r},"
            ")"
        )


class Point(_Geometry):
    """
    A geographic location defined by longitude, latitude, and (optional) altitude.

    When a Point is contained by a Placemark, the point itself determines the position
    of the Placemark's name and icon.
    When a Point is extruded, it is connected to the ground with a line.
    This "tether" uses the current LineStyle.

    https://developers.google.com/kml/documentation/kmlreference#point
    """

    extrude: Optional[bool]
    kml_coordinates: Optional[Coordinates]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        extrude: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        geometry: Optional[geo.Point] = None,
        kml_coordinates: Optional[Coordinates] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Point object.

        Args:
        ----
            ns (Optional[str]): The namespace for the element.
            name_spaces (Optional[Dict[str, str]]): The namespace dictionary for the
            element.
            id (Optional[str]): The ID of the element.
            target_id (Optional[str]): The target ID of the element.
            extrude (Optional[bool]): Whether to extrude the geometry.
            tessellate (Optional[bool]): Whether to tessellate the geometry.
            altitude_mode (Optional[AltitudeMode]): The altitude mode of the geometry.
            geometry (Optional[geo.Point]): The geometry object.
            kml_coordinates (Optional[Coordinates]): The KML coordinates of the point.
            **kwargs (Any): Additional keyword arguments.

        Raises:
        ------
            GeometryError: If both `geometry` and `kml_coordinates` are provided.

        """
        if geometry is not None and kml_coordinates is not None:
            raise GeometryError(MsgMutualExclusive)
        if kml_coordinates is None:
            kml_coordinates = (
                Coordinates(coords=geometry.coords)  # type: ignore[arg-type]
                if geometry
                else None
            )
        self.kml_coordinates = kml_coordinates
        self.extrude = extrude
        kwargs.pop("tessellate", None)
        super().__init__(
            ns=ns,
            id=id,
            name_spaces=name_spaces,
            target_id=target_id,
            altitude_mode=altitude_mode,
            **kwargs,
        )

    def __repr__(self) -> str:
        """
        Return a string representation of the Point object.

        Returns
        -------
            str: The string representation of the Point object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"extrude={self.extrude!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"kml_coordinates={self.kml_coordinates!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Point object has a valid geometry.

        Returns
        -------
            bool: True if the Point object has a valid geometry, False otherwise.

        """
        return bool(self.geometry)

    def __eq__(self, other: object) -> bool:
        """Check if the Point objects are equal."""
        if isinstance(other, Point):
            return all(
                getattr(self, attr) == getattr(other, attr)
                for attr in (
                    "extrude",
                    "altitude_mode",
                    "geometry",
                    *xml_attrs,
                    *self._get_splat(),
                )
            )
        return super().__eq__(other)

    @property
    def geometry(self) -> Optional[geo.Point]:
        """
        Get the geometry object of the Point.

        Returns
        -------
            Optional[geo.Point]: The geometry object of the Point,
            or None if it doesn't exist.

        """
        if not self.kml_coordinates:
            return None
        try:
            return geo.Point.from_coordinates(self.kml_coordinates.coords)
        except (DimensionError, TypeError):
            return None


registry.register(
    Point,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(bool,),
        attr_name="extrude",
        node_name="extrude",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    Point,
    item=RegistryItem(
        ns_ids=("kml", "gx"),
        classes=(AltitudeMode,),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)
registry.register(
    Point,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(Coordinates,),
        attr_name="kml_coordinates",
        node_name="coordinates",
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class LineString(_Geometry):
    """
    Defines a connected set of line segments.

    Use <LineStyle> to specify the color, color mode, and width of the line.
    When a LineString is extruded, the line is extended to the ground, forming a polygon
    that looks somewhat like a wall or fence.
    For extruded LineStrings, the line itself uses the current LineStyle, and the
    extrusion uses the current PolyStyle.
    See the KML Tutorial for examples of LineStrings (or paths).

    https://developers.google.com/kml/documentation/kmlreference#linestring
    """

    extrude: Optional[bool]
    tessellate: Optional[bool]
    kml_coordinates: Optional[Coordinates]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        extrude: Optional[bool] = None,
        tessellate: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        geometry: Optional[geo.LineString] = None,
        kml_coordinates: Optional[Coordinates] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a LineString object.

        Args:
        ----
            ns (Optional[str]): The namespace of the element.
            name_spaces (Optional[Dict[str, str]]): The namespaces used in the element.
            id (Optional[str]): The ID of the element.
            target_id (Optional[str]): The target ID of the element.
            extrude (Optional[bool]): Whether to extrude the geometry.
            tessellate (Optional[bool]): Whether to tessellate the geometry.
            altitude_mode (Optional[AltitudeMode]): The altitude mode of the geometry.
            geometry (Optional[geo.LineString]): The LineString geometry.
            kml_coordinates (Optional[Coordinates]): The coordinates of the geometry.
            **kwargs (Any): Additional keyword arguments.

        Raises:
        ------
            ValueError: If both `geometry` and `kml_coordinates` are provided.

        """
        if geometry is not None and kml_coordinates is not None:
            raise GeometryError(MsgMutualExclusive)
        if kml_coordinates is None:
            kml_coordinates = Coordinates(coords=geometry.coords) if geometry else None
        self.kml_coordinates = kml_coordinates
        self.extrude = extrude
        self.tessellate = tessellate
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            altitude_mode=altitude_mode,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Create a string (c)representation for LineString."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"extrude={self.extrude!r}, "
            f"tessellate={self.tessellate!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"geometry={self.geometry!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the LineString object is non-empty.

        Returns
        -------
            bool: True if the LineString object is non-empty, False otherwise.

        """
        return bool(self.geometry)

    def __eq__(self, other: object) -> bool:
        """Check if the LineString objects is equal."""
        if isinstance(other, LineString):
            return all(
                getattr(self, attr) == getattr(other, attr)
                for attr in (
                    "extrude",
                    "tessellate",
                    "geometry",
                    *xml_attrs,
                    *self._get_splat(),
                )
            )
        return super().__eq__(other)

    @property
    def geometry(self) -> Optional[geo.LineString]:
        """
        Get the LineString geometry.

        Returns
        -------
            Optional[geo.LineString]: The LineString geometry, or None if it doesn't
            exist.

        """
        if not self.kml_coordinates:
            return None
        try:
            return geo.LineString.from_coordinates(self.kml_coordinates.coords)
        except DimensionError:
            return None


registry.register(
    LineString,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(bool,),
        attr_name="extrude",
        node_name="extrude",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    LineString,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(bool,),
        attr_name="tessellate",
        node_name="tessellate",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    LineString,
    item=RegistryItem(
        ns_ids=("kml", "gx"),
        classes=(AltitudeMode,),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)
registry.register(
    LineString,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(Coordinates,),
        attr_name="kml_coordinates",
        node_name="coordinates",
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class LinearRing(LineString):
    """
    Defines a closed line string, typically the outer boundary of a Polygon.

    Optionally, a LinearRing can also be used as the inner boundary of a Polygon to
    create holes in the Polygon.
    A Polygon can contain multiple <LinearRing> elements used as inner boundaries.

    https://developers.google.com/kml/documentation/kmlreference#linearring
    """

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        extrude: Optional[bool] = None,
        tessellate: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        geometry: Optional[geo.LinearRing] = None,
        kml_coordinates: Optional[Coordinates] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Geometry object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the element.
        name_spaces : dict[str, str], optional
            A dictionary of namespace prefixes and URIs.
        id : str, optional
            The ID of the element.
        target_id : str, optional
            The target ID of the element.
        extrude : bool, optional
            Whether to extrude the geometry.
        tessellate : bool, optional
            Whether to tessellate the geometry.
        altitude_mode : AltitudeMode, optional
            The altitude mode of the geometry.
        geometry : geo.LinearRing, optional
            The geometry object.
        kml_coordinates : Coordinates, optional
            The KML coordinates.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            extrude=extrude,
            tessellate=tessellate,
            altitude_mode=altitude_mode,
            geometry=geometry,
            kml_coordinates=kml_coordinates,
            **kwargs,
        )

    @property
    def geometry(self) -> Optional[geo.LinearRing]:
        """
        Get the geometry of the LinearRing.

        Returns
        -------
            Optional[geo.LinearRing]: The geometry of the LinearRing,
            or None if kml_coordinates is not set or if there is a DimensionError.

        """
        if not self.kml_coordinates:
            return None
        try:
            return cast(
                geo.LinearRing,
                geo.LinearRing.from_coordinates(self.kml_coordinates.coords),
            )
        except DimensionError:
            return None


class BoundaryIs(_XMLObject):
    """
    Represents the inner or outer boundary of a polygon in KML.

    Attributes
    ----------
    kml_geometry : Optional[LinearRing]
        The KML geometry representing the outer boundary.

    """

    _default_nsid = config.KML
    kml_geometry: Optional[LinearRing]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        geometry: Optional[geo.LinearRing] = None,
        kml_geometry: Optional[LinearRing] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Boundary object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the KML element.
        name_spaces : dict, optional
            A dictionary of namespace prefixes and URIs.
        geometry : fastkml.geometry.LinearRing, optional
            The geometry object.
        kml_geometry : fastkml.geometry.LinearRing, optional
            The KML geometry object.
        **kwargs : Any
            Additional keyword arguments.

        Raises
        ------
        GeometryError
            If both `geometry` and `kml_geometry` are provided.

        Notes
        -----
        If `kml_geometry` is not provided, it will be created based on the `geometry`
        parameter.

        """
        if geometry is not None and kml_geometry is not None:
            raise GeometryError(MsgMutualExclusive)
        if kml_geometry is None:
            kml_geometry = (
                LinearRing(ns=ns, name_spaces=name_spaces, geometry=geometry)
                if geometry
                else None
            )
        self.kml_geometry = kml_geometry
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )

    def __bool__(self) -> bool:
        """
        Check if the OuterBoundaryIs object is valid.

        Returns
        -------
        bool
            True if the object has a valid geometry, False otherwise.

        """
        return bool(self.geometry)

    def __repr__(self) -> str:
        """Create a string (c)representation for OuterBoundaryIs."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"kml_geometry={self.kml_geometry!r}, "
            f"**{self._get_splat()},"
            ")"
        )

    @property
    def geometry(self) -> Optional[geo.LinearRing]:
        """
        Get the geometry of the OuterBoundaryIs object.

        Returns
        -------
        Optional[geo.LinearRing]
            The geometry object representing the outer boundary.

        """
        return self.kml_geometry.geometry if self.kml_geometry else None


class OuterBoundaryIs(BoundaryIs):
    """Represents the outer boundary of a polygon in KML."""

    @classmethod
    def get_tag_name(cls) -> str:
        """
        Get the tag name for the OuterBoundaryIs object.

        Returns
        -------
        str
            The tag name.

        """
        return "outerBoundaryIs"


class InnerBoundaryIs(BoundaryIs):
    """Represents the inner boundary of a polygon in KML."""

    @classmethod
    def get_tag_name(cls) -> str:
        """Return the tag name of the element."""
        return "innerBoundaryIs"


registry.register(
    BoundaryIs,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(LinearRing,),
        attr_name="kml_geometry",
        node_name="LinearRing",
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)


class Polygon(_Geometry):
    """
    A Polygon is defined by an outer boundary and 0 or more inner boundaries.

    The boundaries, in turn, are defined by LinearRings.
    When a Polygon is extruded, its boundaries are connected to the ground to form
    additional polygons, which gives the appearance of a building or a box.
    Extruded Polygons use <PolyStyle> for their color, color mode, and fill.

    The <coordinates> for polygons must be specified in counterclockwise order.

    The outer boundary is represented by the `outer_boundary_is` attribute,
    which is an instance of the `OuterBoundaryIs` class.
    The inner boundaries are represented by the `inner_boundary_is` attribute,
    which is an instance of the `InnerBoundaryIs` class.

    The `geometry` property returns a `geo.Polygon` object representing the
    geometry of the Polygon.

    Example usage::

        polygon = Polygon(outer_boundary_is=outer_boundary,
            inner_boundary_is=inner_boundary)
        print(polygon.geometry)


    https://developers.google.com/kml/documentation/kmlreference#polygon
    """

    extrude: Optional[bool]
    tessellate: Optional[bool]
    outer_boundary: Optional[OuterBoundaryIs]
    inner_boundaries: List[InnerBoundaryIs]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        extrude: Optional[bool] = None,
        tessellate: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        outer_boundary: Optional[OuterBoundaryIs] = None,
        inner_boundaries: Optional[Iterable[InnerBoundaryIs]] = None,
        geometry: Optional[geo.Polygon] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Geometry object.

        Parameters
        ----------
        ns : Optional[str]
            The namespace of the element.
        name_spaces : Optional[Dict[str, str]]
            The dictionary of namespace prefixes and URIs.
        id : Optional[str]
            The ID of the element.
        target_id : Optional[str]
            The target ID of the element.
        extrude : Optional[bool]
            The extrude flag of the element.
        tessellate : Optional[bool]
            The tessellate flag of the element.
        altitude_mode : Optional[AltitudeMode]
            The altitude mode of the element.
        outer_boundary : Optional[OuterBoundaryIs]
            The outer boundary of the element.
        inner_boundaries : Optional[Iterable[InnerBoundaryIs]]
            The inner boundaries of the element.
        geometry : Optional[geo.Polygon]
            The geometry object of the element.
        **kwargs : Any
            Additional keyword arguments.

        Raises
        ------
        GeometryError:
            If both outer_boundary_is and geometry are provided.

        Returns
        -------
        None

        """
        if outer_boundary is not None and geometry is not None:
            raise GeometryError(MsgMutualExclusive)
        if geometry is not None:
            outer_boundary = OuterBoundaryIs(geometry=geometry.exterior)
            inner_boundaries = [
                InnerBoundaryIs(geometry=interior) for interior in geometry.interiors
            ]
        self.outer_boundary = outer_boundary
        self.inner_boundaries = list(inner_boundaries) if inner_boundaries else []
        self.extrude = extrude
        self.tessellate = tessellate
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            altitude_mode=altitude_mode,
            **kwargs,
        )

    def __bool__(self) -> bool:
        """
        Return True if the outer boundary is defined, False otherwise.

        Returns
        -------
        bool
            True if the outer boundary is defined, False otherwise.

        """
        return bool(self.outer_boundary)

    @property
    def geometry(self) -> Optional[geo.Polygon]:
        """
        Get the geometry object representing the geometry of the Polygon.

        Returns
        -------
        Optional[geo.Polygon]
            The geometry object representing the geometry of the Polygon.

        """
        if not self.outer_boundary:
            return None
        if not self.inner_boundaries:
            return geo.Polygon.from_linear_rings(
                cast(geo.LinearRing, self.outer_boundary.geometry),
            )
        return geo.Polygon.from_linear_rings(
            cast(geo.LinearRing, self.outer_boundary.geometry),
            *[
                interior.geometry
                for interior in self.inner_boundaries
                if interior.geometry is not None
            ],
        )

    def __repr__(self) -> str:
        """
        Create a string representation for Polygon.

        Returns
        -------
        str
            The string representation for Polygon.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"extrude={self.extrude!r}, "
            f"tessellate={self.tessellate!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"geometry={self.geometry!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __eq__(self, other: object) -> bool:
        """Check if the Polygon objects are equal."""
        if isinstance(other, Polygon):
            return all(
                getattr(self, attr) == getattr(other, attr)
                for attr in (
                    "extrude",
                    "tessellate",
                    "geometry",
                    *xml_attrs,
                    *self._get_splat(),
                )
            )
        return super().__eq__(other)


registry.register(
    Polygon,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(bool,),
        attr_name="extrude",
        node_name="extrude",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    Polygon,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(bool,),
        attr_name="tessellate",
        node_name="tessellate",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
        default=False,
    ),
)
registry.register(
    Polygon,
    item=RegistryItem(
        ns_ids=("kml", "gx"),
        classes=(AltitudeMode,),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)
registry.register(
    Polygon,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(OuterBoundaryIs,),
        attr_name="outer_boundary",
        node_name="outerBoundaryIs",
        get_kwarg=xml_subelement_kwarg,
        set_element=xml_subelement,
    ),
)
registry.register(
    Polygon,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(InnerBoundaryIs,),
        attr_name="inner_boundaries",
        node_name="innerBoundaryIs",
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


def create_multigeometry(
    geometries: Sequence[AnyGeometryType],
) -> Optional[MultiGeometryType]:
    """
    Create a MultiGeometry from a sequence of geometries.

    Args:
    ----
        geometries: Sequence of geometries.

    Returns:
    -------
        MultiGeometry

    """
    geom_types = {geom.geom_type for geom in geometries}
    if not geom_types:
        return None
    if len(geom_types) == 1:
        geom_type = geom_types.pop()
        map_to_geometries = {
            geo.Point.__name__: geo.MultiPoint.from_points,
            geo.LineString.__name__: geo.MultiLineString.from_linestrings,
            geo.Polygon.__name__: geo.MultiPolygon.from_polygons,
        }
        for geometry_name, constructor in map_to_geometries.items():
            if geom_type == geometry_name:
                return constructor(  # type: ignore[operator, no-any-return]
                    *geometries,
                )

    return geo.GeometryCollection(geometries)


class MultiGeometry(_BaseObject):
    """A container for zero or more geometry primitives."""

    kml_geometries: List[Union[Point, LineString, Polygon, LinearRing, Self]]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        extrude: Optional[bool] = None,
        tessellate: Optional[bool] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        kml_geometries: Optional[
            Iterable[Union[Point, LineString, Polygon, LinearRing, Self]]
        ] = None,
        geometry: Optional[MultiGeometryType] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Geometry object.

        Parameters
        ----------
        ns : str, optional
            The namespace for the KML element.
        name_spaces : dict, optional
            A dictionary of namespace prefixes and URIs.
        id : str, optional
            The ID of the KML element.
        target_id : str, optional
            The target ID of the KML element.
        kml_geometries : iterable of Point, LineString, Polygon, LinearRing,
            MultiGeometry
            A collection of KML geometries.
        geometry : MultiGeometryType, optional
            A multi-geometry object.
            Parameters for geometry and kml_geometries are mutually exclusive.
            When geometry is provided, kml_geometries will be created from it and
            you can specify additional parameters like extrude, tessellate, and
            altitude_mode which will be set on the individual geometries.
        extrude : bool, optional
            Specifies whether to extend the geometry to the ground.
            This is not set on the multi-geometry itself, but on the individual
            geometries.
        tessellate : bool, optional
            Specifies whether to allow the geometry to follow the terrain.
            This is not set on the multi-geometry itself, but on the individual
            geometries.
        altitude_mode : AltitudeMode, optional
            The altitude mode of the geometry. This is not set on the multi-geometry
            itself, but on the individual geometries.
        **kwargs : any
            Additional keyword arguments.

        Raises
        ------
        GeometryError
            If both `kml_geometries` and `geometry` are provided.

        Notes
        -----
        - If `geometry` is provided, it will be converted into a collection of KML
            geometries.

        """
        if kml_geometries is not None and geometry is not None:
            raise GeometryError(MsgMutualExclusive)
        if geometry is not None:
            kml_geometries = [
                create_kml_geometry(  # type: ignore[misc]
                    geometry=geom,
                    ns=ns,
                    name_spaces=name_spaces,
                    extrude=extrude,
                    tessellate=tessellate,
                    altitude_mode=altitude_mode,
                )
                for geom in geometry.geoms
            ]
        self.kml_geometries = list(kml_geometries) if kml_geometries else []
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )

    def __bool__(self) -> bool:
        """Return True if the MultiGeometry has a geometry, False otherwise."""
        return bool(self.geometry)

    def __repr__(self) -> str:
        """Return a string representation of the MultiGeometry."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"kml_geometries={self.kml_geometries!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    @property
    def geometry(self) -> Optional[MultiGeometryType]:
        """Return the geometry of the MultiGeometry."""
        return create_multigeometry(
            [geom.geometry for geom in self.kml_geometries if geom.geometry],
        )


registry.register(
    MultiGeometry,
    item=RegistryItem(
        ns_ids=("kml",),
        classes=(Point, LineString, Polygon, LinearRing, MultiGeometry),
        attr_name="kml_geometries",
        node_name="(Point|LineString|Polygon|LinearRing|MultiGeometry)",
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


KMLGeometryType = Union[Point, LineString, Polygon, LinearRing, MultiGeometry]


def _unknown_geometry_type(geometry: Union[GeoType, GeoCollectionType]) -> NoReturn:
    """
    Raise an error for an unknown geometry type.

    Args:
    ----
        geometry: The geometry object.

    Raises:
    ------
        KMLWriteError: If the geometry type is unknown.

    """
    msg = f"Unsupported geometry type {type(geometry)}"  # pragma: no cover
    raise KMLWriteError(msg)  # pragma: no cover


def create_kml_geometry(
    geometry: Union[GeoType, GeoCollectionType],
    *,
    ns: Optional[str] = None,
    name_spaces: Optional[Dict[str, str]] = None,
    id: Optional[str] = None,
    target_id: Optional[str] = None,
    extrude: Optional[bool] = None,
    tessellate: Optional[bool] = None,
    altitude_mode: Optional[AltitudeMode] = None,
) -> KMLGeometryType:
    """
    Create a KML geometry from a geometry object.

    Args:
    ----
        geometry: Geometry object.
        ns: Namespace of the object
        name_spaces: Name spaces of the object
        id: Id of the object
        target_id: Target id of the object
        extrude: Specifies whether to connect the feature to the ground with a line.
        tessellate: Specifies whether to allow the LineString to follow the terrain.
        altitude_mode: Specifies how altitude components in the <coordinates>
                       element are interpreted.

    Returns:
    -------
        KML geometry object.

    """
    _map_to_kml: Dict[
        Union[Type[GeoType], Type[GeoCollectionType]],
        Type[KMLGeometryType],
    ] = {
        geo.Point: Point,
        geo.Polygon: Polygon,
        geo.LinearRing: LinearRing,
        geo.LineString: LineString,
        geo.MultiPoint: MultiGeometry,
        geo.MultiLineString: MultiGeometry,
        geo.MultiPolygon: MultiGeometry,
        geo.GeometryCollection: MultiGeometry,
    }
    geom = shape(geometry)
    for geometry_class, kml_class in _map_to_kml.items():
        if isinstance(geom, geometry_class):
            return kml_class(
                ns=ns,
                name_spaces=name_spaces,
                id=id,
                target_id=target_id,
                extrude=extrude,
                tessellate=tessellate,
                altitude_mode=altitude_mode,
                geometry=geom,  # type: ignore[arg-type]
            )

    _unknown_geometry_type(geometry)  # pragma: no cover



# Copyright (C) 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Mixins for the KML classes."""
import logging
from typing import Optional
from typing import Union

from fastkml.times import KmlDateTime
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp

logger = logging.getLogger(__name__)


class TimeMixin:
    """Mixin for classes that have a time element."""

    times: Optional[Union[TimeSpan, TimeStamp]] = None

    @property
    def time_stamp(self) -> Optional[KmlDateTime]:
        """Return the timestamp."""
        return self.times.timestamp if isinstance(self.times, TimeStamp) else None

    @property
    def begin(self) -> Optional[KmlDateTime]:
        """Return the start time of a time span."""
        return self.times.begin if isinstance(self.times, TimeSpan) else None

    @property
    def end(self) -> Optional[KmlDateTime]:
        """Return the end time of a time span."""
        return self.times.end if isinstance(self.times, TimeSpan) else None



# Copyright (C) 2012 -2022 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
Fastkml is a library to read, write and manipulate kml files.

It aims to keep it simple and fast (using lxml if available).
Fast refers to the time you spend to write and read KML files as well as the time
you spend to get acquainted to the library or to create KML objects.
It provides a subset of KML and is aimed at documents that can be read from
multiple clients such as openlayers and google maps rather than to give you all
functionality that KML on google earth provides.
"""
from fastkml.about import __version__  # noqa: F401
from fastkml.atom import Author as AtomAuthor
from fastkml.atom import Contributor as AtomContributor
from fastkml.atom import Link as AtomLink
from fastkml.containers import Document
from fastkml.containers import Folder
from fastkml.data import Data
from fastkml.data import ExtendedData
from fastkml.data import Schema
from fastkml.data import SchemaData
from fastkml.features import Placemark
from fastkml.geometry import LinearRing
from fastkml.geometry import LineString
from fastkml.geometry import MultiGeometry
from fastkml.geometry import Point
from fastkml.geometry import Polygon
from fastkml.kml import KML
from fastkml.links import Icon
from fastkml.links import Link
from fastkml.overlays import GroundOverlay
from fastkml.overlays import PhotoOverlay
from fastkml.styles import BalloonStyle
from fastkml.styles import IconStyle
from fastkml.styles import LabelStyle
from fastkml.styles import LineStyle
from fastkml.styles import PolyStyle
from fastkml.styles import Style
from fastkml.styles import StyleMap
from fastkml.styles import StyleUrl
from fastkml.times import TimeSpan
from fastkml.times import TimeStamp
from fastkml.views import Camera
from fastkml.views import LookAt

__all__ = [
    "KML",
    "Document",
    "Folder",
    "GroundOverlay",
    "Placemark",
    "TimeSpan",
    "TimeStamp",
    "ExtendedData",
    "Data",
    "PhotoOverlay",
    "Schema",
    "SchemaData",
    "StyleUrl",
    "Style",
    "StyleMap",
    "IconStyle",
    "LineStyle",
    "PolyStyle",
    "LabelStyle",
    "BalloonStyle",
    "AtomLink",
    "Icon",
    "Link",
    "Point",
    "LineString",
    "LinearRing",
    "Polygon",
    "MultiGeometry",
    "AtomAuthor",
    "AtomContributor",
    "Camera",
    "LookAt",
]



# Copyright (C) 2012-2023 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Link and Icon elements."""

from typing import Any
from typing import Dict
from typing import Optional

from fastkml.enums import RefreshMode
from fastkml.enums import ViewRefreshMode
from fastkml.helpers import enum_subelement
from fastkml.helpers import float_subelement
from fastkml.helpers import subelement_enum_kwarg
from fastkml.helpers import subelement_float_kwarg
from fastkml.helpers import subelement_text_kwarg
from fastkml.helpers import text_subelement
from fastkml.kml_base import _BaseObject
from fastkml.registry import RegistryItem
from fastkml.registry import registry


class Link(_BaseObject):
    """
    Represents a <Link> element.

    It specifies the location of any of the following:

    - KML files fetched by network links
    - Image files used in any Overlay
    - Model files used in the <Model> element

    https://developers.google.com/kml/documentation/kmlreference#link
    """

    href: str
    refresh_mode: Optional[RefreshMode]
    refresh_interval: Optional[float]
    view_refresh_mode: Optional[ViewRefreshMode]
    view_refresh_time: Optional[float]
    view_bound_scale: Optional[float]
    view_format: str
    http_query: str

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        href: Optional[str] = None,
        refresh_mode: Optional[RefreshMode] = None,
        refresh_interval: Optional[float] = None,
        view_refresh_mode: Optional[ViewRefreshMode] = None,
        view_refresh_time: Optional[float] = None,
        view_bound_scale: Optional[float] = None,
        view_format: Optional[str] = None,
        http_query: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the KML Icon Object."""
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.href = href.strip() if href else ""
        self.refresh_mode = refresh_mode
        self.refresh_interval = refresh_interval
        self.view_refresh_mode = view_refresh_mode
        self.view_refresh_time = view_refresh_time
        self.view_bound_scale = view_bound_scale
        self.view_format = view_format or ""
        self.http_query = http_query or ""

    def __repr__(self) -> str:
        """Create a string (c)representation for Link."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"href={self.href!r}, "
            f"refresh_mode={self.refresh_mode}, "
            f"refresh_interval={self.refresh_interval!r}, "
            f"view_refresh_mode={self.view_refresh_mode}, "
            f"view_refresh_time={self.view_refresh_time!r}, "
            f"view_bound_scale={self.view_bound_scale!r}, "
            f"view_format={self.view_format!r}, "
            f"http_query={self.http_query!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the link has a valid href.

        Returns
        -------
        bool
            True if the link has a valid href, False otherwise.

        """
        return bool(self.href)


registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="href",
        node_name="href",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="refresh_mode",
        node_name="refreshMode",
        classes=(RefreshMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=RefreshMode.on_change,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="refresh_interval",
        node_name="refreshInterval",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=4.0,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="view_refresh_mode",
        node_name="viewRefreshMode",
        classes=(ViewRefreshMode,),
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=ViewRefreshMode.never,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="view_refresh_time",
        node_name="viewRefreshTime",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=4.0,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="view_bound_scale",
        node_name="viewBoundScale",
        classes=(float,),
        get_kwarg=subelement_float_kwarg,
        set_element=float_subelement,
        default=1.0,
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="view_format",
        node_name="viewFormat",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
        default="BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]",
    ),
)
registry.register(
    Link,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="http_query",
        node_name="httpQuery",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)


class Icon(Link):
    """
    Represents an <Icon> element used in IconStyle and Overlays.

    Defines an image associated with an Icon style or overlay.
    The required <href> child element defines the location
    of the image to be used as the overlay or as the icon for the placemark.
    This location can either be on a local file system or a remote web server.

    https://developers.google.com/kml/documentation/kmlreference#icon

    Todo:
    ----
    The <gx:x>, <gx:y>, <gx:w>, and <gx:h> elements are used to select one
    icon from an image that contains multiple icons
    (often referred to as an icon palette).

    """



"""Fastkml utility functions."""

from typing import Any
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

__all__ = ["find", "find_all", "has_attribute_values"]


def has_attribute_values(obj: object, **kwargs: Any) -> bool:
    """
    Check if an object has all of the given attribute values.

    Args:
    ----
        obj: The object to check.
        **kwargs: Attributes of the object to match.

    Returns:
    -------
        True if the object has the given attribute values, False otherwise.

    """
    try:
        return all(getattr(obj, key) == value for key, value in kwargs.items())
    except AttributeError:
        return False


def find_all(
    obj: object,
    *,
    of_type: Optional[Union[Type[object], Tuple[Type[object], ...]]] = None,
    **kwargs: Any,
) -> Generator[object, None, None]:
    """
    Find all instances of a given type in a given object.

    Args:
    ----
        obj: The object to search.
        of_type: The type(s) to search for or None for any type.
        **kwargs: Attributes of the object to match.

    Returns:
    -------
        An iterable of all instances of the given type in the given object.

    """
    if (of_type is None or isinstance(obj, of_type)) and has_attribute_values(
        obj,
        **kwargs,
    ):
        yield obj
    try:
        attrs = (attr for attr in obj.__dict__ if not attr.startswith("_"))
    except AttributeError:
        return
    for attr_name in attrs:
        attr = getattr(obj, attr_name)
        try:
            for item in attr:
                yield from find_all(item, of_type=of_type, **kwargs)
        except TypeError:
            yield from find_all(attr, of_type=of_type, **kwargs)


def find(
    obj: object,
    *,
    of_type: Optional[Union[Type[object], Tuple[Type[object], ...]]] = None,
    **kwargs: Any,
) -> Optional[object]:
    """
    Find the first instance of a given type in a given object.

    Args:
    ----
        obj: The object to search.
        of_type: The type(s) to search for or None for any type.
        **kwargs: Attributes of the object to match.

    Returns:
    -------
        The first instance of the given type in the given object or None if not found.

    """
    return next(find_all(obj, of_type=of_type, **kwargs), None)



# Copyright (C) 2022 - 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""
Add Custom Data.

https://developers.google.com/kml/documentation/extendeddata
"""

import logging
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from fastkml.base import _XMLObject
from fastkml.enums import DataType
from fastkml.exceptions import KMLSchemaError
from fastkml.helpers import attribute_enum_kwarg
from fastkml.helpers import attribute_text_kwarg
from fastkml.helpers import enum_attribute
from fastkml.helpers import node_text
from fastkml.helpers import node_text_kwarg
from fastkml.helpers import subelement_text_kwarg
from fastkml.helpers import text_attribute
from fastkml.helpers import text_subelement
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.registry import RegistryItem
from fastkml.registry import registry

__all__ = [
    "Data",
    "ExtendedData",
    "Schema",
    "SchemaData",
    "SimpleField",
]

logger = logging.getLogger(__name__)


class SimpleField(_XMLObject):
    """
    A SimpleField always has both name and type attributes.

    The declaration of the custom field, must specify both the type
    and the name of this field.
    If either the type or the name is omitted, the field is ignored.

    The type can be one of the following:
     - string
     - int
     - uint
     - short
     - ushort
     - float
     - double
     - bool

    The displayName, if any, to be used when the field name is displayed to
    the Google Earth user. Use the [CDATA] element to escape standard
    HTML markup.
    """

    _default_nsid = "kml"

    name: Optional[str]
    type_: Optional[DataType]
    display_name: Optional[str]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        type_: Optional[DataType] = None,
        display_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Data class.

        Args:
        ----
            ns (Optional[str]): The namespace of the data.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            name (Optional[str]): The name of the data.
            type_ (Optional[DataType]): The type of the data.
            display_name (Optional[str]): The display name of the data.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.name = name.strip() or None if name else None
        self.type_ = type_ or None
        self.display_name = display_name.strip() or None if display_name else None

    def __repr__(self) -> str:
        """
        Return a string representation of the SimpleField object.

        Returns
        -------
            str: A string representation of the SimpleField object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"name={self.name!r}, "
            f"type_={self.type_}, "
            f"display_name={self.display_name!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the object is considered True or False.

        Returns
        -------
            bool: True if both the name and type are non-empty, False otherwise.

        """
        return bool(self.name) and bool(self.type_)


registry.register(
    SimpleField,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="display_name",
        node_name="displayName",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    SimpleField,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="name",
        node_name="name",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)
registry.register(
    SimpleField,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="type_",
        node_name="type",
        classes=(DataType,),
        get_kwarg=attribute_enum_kwarg,
        set_element=enum_attribute,
    ),
)


class Schema(_XMLObject):
    """
    Specifies a custom KML schema that is used to add custom data to KML Features.

    The "id" attribute is required and must be unique within the KML file.
    <Schema> is always a child of <Document>.

    https://developers.google.com/kml/documentation/extendeddata#declare-the-schema-element

    """

    _default_nsid = "kml"

    name: Optional[str]
    fields: List[SimpleField]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
        fields: Optional[Iterable[SimpleField]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a Schema object.

        Parameters
        ----------
        ns : str, optional
            The namespace of the schema.
        name_spaces : dict[str, str], optional
            The dictionary of namespace prefixes and URIs.
        id : str, optional
            The unique identifier for the schema.
        target_id : str, optional
            The target identifier for the schema.
        name : str, optional
            The name of the schema.
        fields : Iterable[SimpleField], optional
            The list of fields in the schema.
        **kwargs : Any
            Additional keyword arguments.

        Raises
        ------
        KMLSchemaError
            If the id is not provided.

        """
        if not id:
            msg = "Id is required for schema"
            raise KMLSchemaError(msg)
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.name = name.strip() or None if name else None
        self.fields = list(fields) if fields else []
        self.id = id.strip() or None if id else None

    def __repr__(self) -> str:
        """
        Return a string representation of the Schema object.

        Returns
        -------
        str
            The string representation of the Schema object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"name={self.name!r}, "
            f"fields={self.fields!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def append(self, field: SimpleField) -> None:
        """
        Append a field to the schema.

        Parameters
        ----------
        field : SimpleField
            The field to be appended.

        """
        self.fields.append(field)


registry.register(
    Schema,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="id",
        node_name="id",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)
registry.register(
    Schema,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="name",
        node_name="name",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)
registry.register(
    Schema,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="fields",
        node_name="SimpleField",
        classes=(SimpleField,),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


class Data(_BaseObject):
    """Represents an untyped name/value pair with optional display name."""

    name: Optional[str]
    value: Optional[str]
    display_name: Optional[str]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        name: Optional[str] = None,
        value: Optional[str] = None,
        display_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Data class.

        Args:
        ----
            ns (Optional[str]): The namespace of the data.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            id (Optional[str]): The ID of the data.
            target_id (Optional[str]): The target ID of the data.
            name (Optional[str]): The name of the data.
            value (Optional[str]): The value of the data.
            display_name (Optional[str]): The display name of the data.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.name = name.strip() or None if name else None
        self.value = value.strip() or None if value else None
        self.display_name = display_name.strip() or None if display_name else None

    def __repr__(self) -> str:
        """
        Create a string representation for Data.

        Returns
        -------
        str
            The string representation of the Data object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"name={self.name!r}, "
            f"value={self.value!r}, "
            f"display_name={self.display_name!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the Data object is truthy.

        Returns
        -------
        bool
            True if the Data object has a name and a non-None value, False otherwise.

        """
        return bool(self.name) and bool(self.value)


registry.register(
    Data,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="display_name",
        node_name="displayName",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    Data,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="value",
        node_name="value",
        classes=(str,),
        get_kwarg=subelement_text_kwarg,
        set_element=text_subelement,
    ),
)
registry.register(
    Data,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="name",
        node_name="name",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)


class SimpleData(_XMLObject):
    _default_nsid = "kml"

    name: Optional[str]
    value: Optional[str]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        name: Optional[str] = None,
        value: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a SimpleData object.

        Args:
        ----
            ns (Optional[str]): The namespace of the object.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            name (Optional[str]): The name of the object.
            value (Optional[str]): The value of the object.
            **kwargs: Additional keyword arguments.

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.name = name.strip() or None if name else None
        self.value = value.strip() or None if value else None

    def __repr__(self) -> str:
        """
        Return a string representation of the SimpleData object.

        Returns
        -------
            str: The string representation of the SimpleData object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"name={self.name!r}, "
            f"value={self.value!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the SimpleData object is truthy.

        Returns
        -------
            bool: True if the name and the value attribute are set, False otherwise.

        """
        return bool(self.name) and bool(self.value)


registry.register(
    SimpleData,
    RegistryItem(
        ns_ids=("kml", ""),
        attr_name="value",
        node_name="value",
        classes=(str,),
        get_kwarg=node_text_kwarg,
        set_element=node_text,
    ),
)
registry.register(
    SimpleData,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="name",
        node_name="name",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)


class SchemaData(_BaseObject):
    """
    Represents the SchemaData element in KML.

    The SchemaData element is used in conjunction with Schema to add typed
    custom data to a KML Feature. The Schema element (identified by the
    schemaUrl attribute) declares the custom data type. The actual data
    objects ("instances" of the custom data) are defined using the
    SchemaData element.
    """

    schema_url: Optional[str]
    data: List[SimpleData]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        schema_url: Optional[str] = None,
        data: Optional[Iterable[SimpleData]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Data class.

        Args:
        ----
            ns (Optional[str]): The namespace for the data.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            id (Optional[str]): The ID of the data.
            target_id (Optional[str]): The target ID of the data.
            schema_url (Optional[str]): The URL of the schema for the data.
            data (Optional[Iterable[SimpleData]]): The iterable of SimpleData objects.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.schema_url = schema_url.strip() or None if schema_url else None
        self.data = list(data) if data else []

    def __repr__(self) -> str:
        """Create a string representation for SchemaData."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"schema_url={self.schema_url!r}, "
            f"data={self.data!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the SchemaData object contains data.

        Returns
        -------
            bool: True if the SchemaData object contains data and a
                schema URL, False otherwise.

        """
        return bool(self.data) and bool(self.schema_url)

    def append_data(self, data: SimpleData) -> None:
        """
        Append a SimpleData object to the SchemaData.

        Args:
        ----
            data (SimpleData): The SimpleData object to be appended.

        """
        self.data.append(data)


registry.register(
    SchemaData,
    RegistryItem(
        ns_ids=("", "kml"),
        attr_name="schema_url",
        node_name="schemaUrl",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)
registry.register(
    SchemaData,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="data",
        node_name="SimpleData",
        classes=(SimpleData,),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)


class ExtendedData(_XMLObject):
    """Represents a list of untyped name/value pairs."""

    _default_nsid = "kml"
    elements: List[Union[Data, SchemaData]]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        elements: Optional[Iterable[Union[Data, SchemaData]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Data class.

        Args:
        ----
            ns (Optional[str]): The namespace for the data.
            name_spaces (Optional[Dict[str, str]]):
                The dictionary of namespace prefixes and URIs.
            elements (Optional[Iterable[Union[Data, SchemaData]]]):
                The iterable of data elements.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            - None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            **kwargs,
        )
        self.elements = [e for e in elements if e] if elements else []

    def __repr__(self) -> str:
        """
        Return a string representation of the ExtendedData object.

        Returns
        -------
            str: A string representation of the ExtendedData object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"elements={self.elements!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """
        Check if the object has any elements.

        Returns
        -------
        bool
            True if the object has elements, False otherwise.

        """
        return bool(self.elements)


registry.register(
    ExtendedData,
    RegistryItem(
        ns_ids=("kml",),
        attr_name="elements",
        node_name="Data,SchemaData",
        classes=(
            Data,
            SchemaData,
        ),
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)



# Copyright (C) 2012 - 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Abstract base classes."""
from typing import Any
from typing import Dict
from typing import Optional

from fastkml import config
from fastkml.base import _XMLObject
from fastkml.helpers import attribute_text_kwarg
from fastkml.helpers import text_attribute
from fastkml.registry import RegistryItem
from fastkml.registry import registry

__all__ = ["_BaseObject"]


class _BaseObject(_XMLObject):
    """
    Base class for all KML objects.

    This is an abstract base class and cannot be used directly in a
    KML file. It provides the id attribute, which allows unique
    identification of a KML element, and the targetId attribute,
    which is used to reference objects that have already been loaded into
    Google Earth. The id attribute must be assigned if the <Update>
    mechanism is to be used.
    """

    _default_nsid = config.KML

    id = None
    target_id = None

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the KML Object.

        Parameters
        ----------
        ns: (Optional[str])
            The namespace of the KML object.
        name_spaces: (Optional[Dict[str, str]])
            The dictionary of namespace prefixes and URIs.
        id: (Optional[str])
            The id attribute of the KML object.
        target_id: (Optional[str])
            The targetId attribute of the KML object.
        kwargs: (Any)
            Additional keyword arguments.

        Returns
        -------
        None

        """
        super().__init__(ns=ns, name_spaces=name_spaces, **kwargs)
        self.id = id or ""
        self.target_id = target_id or ""

    def __repr__(self) -> str:
        """
        Create a string representation for _BaseObject.

        Returns
        -------
        str: The string representation of _BaseObject.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )


registry.register(
    _BaseObject,
    item=RegistryItem(
        ns_ids=("", "kml"),
        attr_name="id",
        node_name="id",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)

registry.register(
    _BaseObject,
    item=RegistryItem(
        ns_ids=("", "kml"),
        attr_name="target_id",
        node_name="targetId",
        classes=(str,),
        get_kwarg=attribute_text_kwarg,
        set_element=text_attribute,
    ),
)



# Copyright (C) 2023 - 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""Helper functions for fastkml."""

import logging
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

from pygeoif.types import PointType

from fastkml import config
from fastkml.enums import Verbosity
from fastkml.exceptions import KMLParseError
from fastkml.types import Element

if TYPE_CHECKING:
    from fastkml.base import _XMLObject
    from fastkml.times import KmlDateTime


logger = logging.getLogger(__name__)


def handle_error(
    *,
    error: Exception,
    strict: bool,
    element: Element,
    node: Element,
    expected: str,
) -> None:
    """
    Handle an error.

    Args:
    ----
        error (Exception): The exception that occurred.
        strict (bool): A flag indicating whether to raise an exception or log a warning.
        element (Element): The XML element being parsed.
        node (Element): The XML node that caused the error.
        expected (str): The expected format or value.

    Raises:
    ------
        KMLParseError: If `strict` is True, the function raises a `KMLParseError` with
        the error message.

    """
    serialized_element = config.etree.tostring(
        element,
        encoding="UTF-8",
    ).decode(
        "UTF-8",
    )
    serialized_node = config.etree.tostring(
        node,
        encoding="UTF-8",
    ).decode(
        "UTF-8",
    )
    msg = (
        f"Error parsing '{serialized_node}' in '{serialized_element}'; "
        f"expected: {expected}. \n {error}"
    )
    if strict:
        raise KMLParseError(msg) from error
    logger.warning("%s, %s", error, msg)


def get_ns(obj: "_XMLObject", value: object) -> str:
    """Get the namespace of an attribute, fall back on the objects namespace."""
    try:
        return obj.name_spaces.get(value.get_ns_id(), "")  # type: ignore[attr-defined]
    except AttributeError:
        return obj.ns


def get_value(
    obj: "_XMLObject",
    *,
    attr_name: str,
    verbosity: Verbosity,
    default: Optional[Any],
) -> Optional[Any]:
    """
    Get the value of an attribute from an object.

    If the verbosity is set to `Verbosity.terse`, the function returns `None` if the
    attribute value is equal to the default value. If the verbosity is set to
    `Verbosity.verbose`, the function returns the default value if the attribute value
    is `None`.

    Args:
    ----
        obj ("_XMLObject"): The object to get the attribute value from.
        attr_name (str): The name of the attribute to retrieve.
        verbosity (Optional[Verbosity]): The verbosity.
        default (Optional[Any]): The default value.

    """
    value = getattr(obj, attr_name, None)
    if value is None and default is not None and verbosity == Verbosity.verbose:
        return default
    return None if value == default and verbosity == Verbosity.terse else value


def node_text(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """
    Set the text of an XML element based on the attribute value in the given object.

    Parameters
    ----------
    obj : "_XMLObject"
        The object containing the attribute value.
    element : Element
        The XML element to set the text content for.
    attr_name : str
        The name of the attribute in the object.
    node_name : str
        The name of the XML node (unused).
    precision : Optional[int]
        The precision to use when converting numeric values to text (unused).
    verbosity : Optional[Verbosity]
        The verbosity level for logging (unused).
    default : Optional[str]
        The default value for the attribute.

    Returns
    -------
    None
        This function does not return anything.

    """
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        element.text = value


def text_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """
    Set the value of an attribute from a subelement with a text node.

    Args:
    ----
        obj ("_XMLObject"): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the subelement to create.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Optional[str]): The default value for the attribute.

    Returns:
    -------
        None

    """
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        subelement = config.etree.SubElement(
            element,
            f"{obj.ns}{node_name}",
        )
        subelement.text = value


def text_attribute(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """
    Set the value of an attribute from a subelement with a text node.

    Args:
    ----
        obj ("_XMLObject"): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the attribute to be set.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Optional[str]): The default value for the attribute.

    Returns:
    -------
        None

    """
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        element.set(node_name, value)


def bool_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[bool],
) -> None:
    """
    Set the value of an attribute from a subelement with a text node.

    Args:
    ----
        obj ("_XMLObject"): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the subelement to create.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Optional[bool]): The default value for the attribute.

    Returns:
    -------
        None

    """
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        subelement = config.etree.SubElement(
            element,
            f"{obj.ns}{node_name}",
        )
        subelement.text = str(int(value))


def int_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[int],
) -> None:
    """
    Set the value of an attribute from a subelement with a text node.

    Args:
    ----
        obj ("_XMLObject"): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the subelement to create.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Optional[int]): The default value for the attribute.

    Returns:
    -------
        None: This function does not return anything.

    """
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        subelement = config.etree.SubElement(
            element,
            f"{obj.ns}{node_name}",
        )
        subelement.text = str(value)


def int_attribute(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[int],
) -> None:
    """
    Set the value of an attribute.

    Args:
    ----
        obj ("_XMLObject"): The object from which to retrieve the attribute value.
        element (Element): The parent element to add the subelement to.
        attr_name (str): The name of the attribute to retrieve the value from.
        node_name (str): The name of the attribute to be set.
        precision (Optional[int]): The precision of the attribute value.
        verbosity (Optional[Verbosity]): The verbosity level.
        default (Optional[int]): The default value for the attribute.

    Returns:
    -------
        None: This function does not return anything.

    """
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        element.set(node_name, str(value))


def float_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[float],
) -> None:
    """Set the value of an attribute from a subelement with a text node."""
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        subelement = config.etree.SubElement(
            element,
            f"{obj.ns}{node_name}",
        )
        subelement.text = str(value)


def float_attribute(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[float],
) -> None:
    """Set the value of an attribute."""
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        element.set(node_name, str(value))


def enum_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[Enum],
) -> None:
    """Set the value of an attribute from a subelement with a text node."""
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        ns = get_ns(obj, value)
        subelement = config.etree.SubElement(
            element,
            f"{ns or ''}{node_name}",
        )
        subelement.text = value.value


def enum_attribute(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[Enum],
) -> None:
    """Set the value of an attribute."""
    value = get_value(obj, attr_name=attr_name, verbosity=verbosity, default=default)
    if value is not None:
        element.set(node_name, value.value)


def datetime_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """Create the subelement for a KML datetime values."""
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        ns = get_ns(obj, value)
        subelement = config.etree.SubElement(
            element,
            f"{ns}{node_name}",
        )
        subelement.text = str(value)


def datetime_subelement_list(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """Create the subelements for a list of KML datetime values."""
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        for item in value:
            ns = get_ns(obj, item)
            subelement = config.etree.SubElement(
                element,
                f"{ns}{node_name}",
            )
            subelement.text = str(item)


def coords_subelement_list(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[str],
) -> None:
    """Create the subelements for a list of KML coordinate values."""
    if value := get_value(
        obj,
        attr_name=attr_name,
        verbosity=verbosity,
        default=default,
    ):
        for coord in value:
            ns = get_ns(obj, coord)
            subelement = config.etree.SubElement(
                element,
                f"{ns}{node_name}",
            )
            if precision is None:
                subelement.text = " ".join(str(c) for c in coord)
            else:
                subelement.text = " ".join(f"{c:.{precision}f}" for c in coord)


def xml_subelement(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional["_XMLObject"],
) -> None:
    """
    Add a subelement to an XML element based on the value of an attribute of an object.

    Args:
    ----
        obj ("_XMLObject"): The object containing the attribute.
        element (Element): The XML element to which the subelement will be added.
        attr_name (str): The name of the attribute in the object.
        node_name (str): The name of the XML node for the subelement (unused).
        precision (Optional[int]): The precision for formatting numerical values.
        verbosity (Optional[Verbosity]): The verbosity level for the subelement.
        default (Optional["_XMLObject"]): The default value for the attribute (unused).

    Returns:
    -------
        None

    """
    if getattr(obj, attr_name, None):
        element.append(
            getattr(obj, attr_name).etree_element(
                precision=precision,
                verbosity=verbosity,
            ),
        )


def xml_subelement_list(
    obj: "_XMLObject",
    *,
    element: Element,
    attr_name: str,
    node_name: str,
    precision: Optional[int],
    verbosity: Verbosity,
    default: Optional[List["_XMLObject"]],
) -> None:
    """
    Add subelements to an XML element based on a list attribute of an object.

    Args:
    ----
        obj ("_XMLObject"): The object containing the list attribute.
        element (Element): The XML element to which the subelements will be added.
        attr_name (str): The name of the list attribute in the object.
        node_name (str): The name of the XML node for each subelement (unused).
        precision (Optional[int]): The precision for floating-point values.
        verbosity (Optional[Verbosity]): The verbosity level for the XML output.
        default (Optional[List["_XMLObject"]]): The default value for the attribute.

    Returns:
    -------
        None

    """
    if getattr(obj, attr_name, None):
        for item in getattr(obj, attr_name):
            if item:
                element.append(
                    item.etree_element(precision=precision, verbosity=verbosity),
                )


def node_text_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, str]:
    """
    Extract the text content of an XML element and return it as a dictionary.

    Args:
    ----
        element (Element): The XML element to extract the text content from.
        ns (str): The namespace of the XML element.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to their
        URIs.
        node_name (str): The name of the XML node.
        kwarg (str): The name of the keyword argument to store the text content in.
        classes (Tuple[Type[object], ...]): A tuple of known types.
        strict (bool): A flag indicating whether to enforce strict parsing rules.

    Returns:
    -------
        Dict[str, str]: A dictionary containing the keyword argument and its
            corresponding text content, if it exists.

    """
    return (
        {kwarg: element.text.strip()} if element.text and element.text.strip() else {}
    )


def subelement_text_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, str]:
    """
    Extract the text content of a subelement and return it as a dictionary.

    Args:
    ----
        element (Element): The parent element.
        ns (str): The namespace of the subelement.
        name_spaces (Dict[str, str]): A dictionary of namespace prefixes and URIs.
        node_name (str): The name of the subelement.
        kwarg (str): The key to use in the returned dictionary.
        classes (Tuple[Type[object], ...]): A tuple of known types.
        strict (bool): A flag indicating whether to enforce strict parsing.

    Returns:
    -------
        Dict[str, str]: A dictionary containing the extracted text content,
            with the specified key.

    """
    node = element.find(f"{ns}{node_name}")
    if node is None or node.text is None:
        return {}
    assert isinstance(node.text, str)  # noqa: S101
    return {kwarg: node.text.strip()} if node.text and node.text.strip() else {}


def attribute_text_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, str]:
    """
    Return a dictionary representing the attribute as a keyword argument.

    Args:
    ----
        element (Element): The XML element.
        ns (str): The namespace of the attribute.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the XML node.
        kwarg (str): The name of the keyword argument.
        classes (Tuple[Type[object], ...]): A tuple of known types.
        strict (bool): A flag indicating whether to enforce strict parsing.

    Returns:
    -------
        Dict[str, str]: A dictionary representing the attribute as a keyword argument.

    """
    attr = element.get(f"{ns}{node_name}")
    return {kwarg: attr} if attr else {}


def _get_boolean_value(*, text: str, strict: bool) -> bool:
    if not strict:
        text = text.lower()
    if text in {"1", "true"}:
        return True
    if text in {"0", "false"}:
        return False
    if not strict:
        return bool(float(text))
    msg = f"Value {text} is not a valid value for Boolean"
    raise ValueError(msg)


def subelement_bool_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, bool]:
    """
    Extract a boolean value from a subelement of an XML element.

    Args:
    ----
        element (Element): The XML element to search for the subelement.
        ns (str): The namespace of the subelement.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the subelement.
        kwarg (str): The name of the keyword argument to store the boolean value.
        classes (Tuple[Type[object], ...]): A tuple of known types.
        strict (bool): A flag indicating whether to enforce strict parsing.

    Returns:
    -------
        Dict[str, bool]: A dictionary containing the keyword argument and its value.

    Raises:
    ------
        ValueError: If the value of the subelement is not a valid boolean.

    """
    assert len(classes) == 1  # noqa: S101
    assert issubclass(classes[0], bool)  # noqa: S101
    node = element.find(f"{ns}{node_name}")
    if node is None:
        return {}
    if node.text and node.text.strip():
        try:
            return {kwarg: _get_boolean_value(text=node.text.strip(), strict=strict)}
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=node,
                expected="Boolean",
            )
    return {}


def subelement_int_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, int]:
    """
    Extract an integer value from a subelement of an XML element.

    Args:
    ----
        element (Element): The XML element to search for the subelement.
        ns (str): The namespace of the subelement.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the subelement.
        kwarg (str): The key to use in the returned dictionary.
        classes (Tuple[Type[object], ...]): A tuple of known types for error handling.
        strict (bool): A flag indicating whether to enforce strict parsing.

    Returns:
    -------
        Dict[str, int]: A dictionary containing the keyword argument and its value.

    Raises:
    ------
        ValueError: If the value of the subelement is not a valid integer and strict.

    """
    node = element.find(f"{ns}{node_name}")
    if node is None:
        return {}
    if node.text and node.text.strip():
        try:
            return {kwarg: int(node.text.strip())}
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=node,
                expected="Integer",
            )
    return {}


def attribute_int_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, int]:
    """
    Extract an integer attribute from an XML element and return it as a dictionary.

    Args:
    ----
        element (Element): The XML element to extract the attribute from.
        ns (str): The namespace of the attribute.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the XML node containing the attribute.
        kwarg (str): The name of the keyword argument to store the extracted attribute.
        classes (Tuple[Type[object], ...]): A tuple of known types (unused).
        strict (bool): A flag indicating whether to raise an exception (unused).

    Returns:
    -------
        Dict[str, int]: A dictionary containing the extracted attribute value.

    """
    attr = element.get(f"{ns}{node_name}")
    return {kwarg: int(attr)} if attr else {}


def subelement_float_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, float]:
    """
    Extract a float value from a subelement of an XML element.

    Args:
    ----
        element (Element): The XML element to search for the subelement.
        ns (str): The namespace of the subelement.
        name_spaces (Dict[str, str]): A dictionary of namespace prefixes and URIs.
        node_name (str): The name of the subelement.
        kwarg (str): The name of the keyword argument to store the float value.
        classes (Tuple[Type[object], ...]): A tuple of known types for error handling.
        strict (bool): A flag indicating whether to raise an error.

    Returns:
    -------
        Dict[str, float]: A dictionary containing the float value as a keyword argument.

    Raises:
    ------
        ValueError: If the value of the subelement cannot be converted and strict.

    """
    node = element.find(f"{ns}{node_name}")
    if node is None:
        return {}
    if node.text and node.text.strip():
        try:
            return {kwarg: float(node.text.strip())}
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=node,
                expected="Float",
            )
    return {}


def attribute_float_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, float]:
    """
    Convert an attribute value to a float and return it as a dictionary.

    Args:
    ----
        element (Element): The XML element containing the attribute.
        ns (str): The namespace of the attribute.
        name_spaces (Dict[str, str]): A dictionary of namespace prefixes and URIs.
        node_name (str): The name of the attribute.
        kwarg (str): The name of the keyword argument to store the converted float.
        classes (Tuple[Type[object], ...]): A tuple of known types for error handling.
        strict (bool): A flag indicating whether to raise an error for invalid values.

    Returns:
    -------
        Dict[str, float]: A dictionary containing the float as a keyword argument.

    Raises:
    ------
        ValueError: If the attribute value cannot be converted to a float.

    """
    attr = element.get(f"{ns}{node_name}")
    try:
        return {kwarg: float(attr)} if attr else {}
    except ValueError as exc:
        handle_error(
            error=exc,
            strict=strict,
            element=element,
            node=element,
            expected="Float",
        )
    return {}


def _get_enum_value(*, enum_class: Type[Enum], text: str, strict: bool) -> Enum:
    value = enum_class(text)
    if strict and value.value != text:
        msg = f"Value {text} is not a valid value for Enum {enum_class.__name__}"
        raise ValueError(msg)
    return value


def subelement_enum_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, Enum]:
    """
    Extract an enumerated value from a subelement of an XML element.

    Args:
    ----
        element (Element): The XML element to search for the subelement.
        ns (str): The namespace of the subelement.
        name_spaces (Dict[str, str]): A dictionary of namespace prefixes and URIs.
        node_name (str): The name of the subelement.
        kwarg (str): The name of the keyword argument to store the extracted value.
        classes (Tuple[Type[object], ...]): A tuple of enumerated value classes.
        strict (bool): A flag indicating whether to raise an exception.

    Returns:
    -------
        Dict[str, Enum]: A dictionary containing the extracted enumerated value.

    Raises:
    ------
        ValueError: If the extracted value is not a valid enumerated value and strict.

    """
    assert len(classes) == 1  # noqa: S101
    assert issubclass(classes[0], Enum)  # noqa: S101
    node = element.find(f"{ns}{node_name}")
    if node is None:
        return {}
    node_text = node.text.strip() if node.text else ""
    if node_text:
        try:
            return {
                kwarg: _get_enum_value(
                    enum_class=classes[0],
                    text=node_text,
                    strict=strict,
                ),
            }
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=node,
                expected="Enum",
            )
    return {}


def attribute_enum_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, Enum]:
    """
    Return a dictionary with the specified keyword argument and its enum value.

    Args:
    ----
        element (Element): The XML element.
        ns (str): The namespace of the XML element.
        name_spaces (Dict[str, str]): A dictionary of namespace prefixes and their URIs.
        node_name (str): The name of the XML node.
        kwarg (str): The name of the keyword argument.
        classes (Tuple[Type[object], ...]): A tuple of enum classes.
        strict (bool): A flag indicating whether to raise an error for invalid values.

    Returns:
    -------
        Dict[str, Enum]: A dictionary with the specified keyword argument and its value.

    """
    assert len(classes) == 1  # noqa: S101
    assert issubclass(classes[0], Enum)  # noqa: S101
    if raw := element.get(f"{ns}{node_name}"):
        try:
            return {
                kwarg: _get_enum_value(
                    enum_class=classes[0],
                    text=raw,
                    strict=strict,
                ),
            }
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=element,
                expected="Enum",
            )
    return {}


def datetime_subelement_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, List["KmlDateTime"]]:
    """Extract a KML datetime from a subelement of an XML element."""
    cls = classes[0]
    node = element.find(f"{ns}{node_name}")
    if node is None:
        return {}
    node_text = node.text.strip() if node.text else ""
    if node_text:
        try:
            return {kwarg: cls.parse(node_text)}  # type: ignore[attr-defined]
        except ValueError as exc:
            handle_error(
                error=exc,
                strict=strict,
                element=element,
                node=node,
                expected="DateTime",
            )
    return {}


def datetime_subelement_list_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, List["KmlDateTime"]]:
    """Extract a list of KML datetime values from subelements of an XML element."""
    args_list: List[KmlDateTime] = []
    cls = classes[0]
    if subelements := element.findall(f"{ns}{node_name}"):
        for subelement in subelements:
            try:
                args_list.append(
                    cls.parse(subelement.text),  # type: ignore[attr-defined]
                )
            except ValueError as exc:  # noqa: PERF203
                handle_error(
                    error=exc,
                    strict=strict,
                    element=element,
                    node=subelement,
                    expected="DateTime",
                )
    return {kwarg: args_list} if args_list else {}


def coords_subelement_list_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, List[PointType]]:
    """Extract a list of KML coordinate values from subelements of an XML element."""
    args_list: List[PointType] = []
    if subelements := element.findall(f"{ns}{node_name}"):
        for subelement in subelements:
            if subelement.text:
                try:
                    coords = cast(
                        PointType,
                        tuple(float(coord) for coord in subelement.text.split()),
                    )
                    args_list.append(coords)
                except ValueError as exc:
                    handle_error(
                        error=exc,
                        strict=strict,
                        element=element,
                        node=subelement,
                        expected="Coordinates",
                    )
    return {kwarg: args_list} if args_list else {}


def xml_subelement_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, "_XMLObject"]:
    """
    Return the subelement of the given XML element based on the provided parameters.

    Args:
    ----
    element (Element): The XML element to search within.
    ns (str): The namespace of the XML element.
    name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to their URIs.
    node_name (str): The name of the XML node to search for.
    kwarg (str): The name of the keyword argument to store the found subelement.
    classes (Tuple[Type[object], ...]): A tuple of classes that represent the types.
    strict (bool): A flag indicating whether to enforce strict parsing rules.

    Returns:
    -------
    Dict[str, "_XMLObject"]: A dictionary containing the found subelement as the value
        of the specified keyword argument.

    """
    for cls in classes:
        subelement = element.find(
            f"{ns}{cls.get_tag_name()}",  # type: ignore[attr-defined]
        )
        if subelement is not None:
            return {
                kwarg: cls.class_from_element(  # type: ignore[attr-defined]
                    ns=ns,
                    name_spaces=name_spaces,
                    element=subelement,
                    strict=strict,
                ),
            }
    return {}


def xml_subelement_list_kwarg(
    *,
    element: Element,
    ns: str,
    name_spaces: Dict[str, str],
    node_name: str,
    kwarg: str,
    classes: Tuple[Type[object], ...],
    strict: bool,
) -> Dict[str, List["_XMLObject"]]:
    """
    Return a dictionary with the specified keyword argument and its list of subelements.

    Args:
    ----
        element (Element): The XML element to search within.
        ns (str): The namespace of the XML element.
        name_spaces (Dict[str, str]): A dictionary mapping namespace prefixes to URIs.
        node_name (str): The name of the XML node to search for.
        kwarg (str): The name of the keyword argument to store the found subelements.
        classes (Tuple[Type[object], ...]): A tuple of classes that represent the types.
        strict (bool): A flag indicating whether to enforce strict parsing rules.

    Returns:
    -------
        Dict[str, List["_XMLObject"]]: A dictionary containing the specified keyword
            argument and its list of subelements.

    """
    args_list = []
    assert node_name is not None  # noqa: S101
    assert name_spaces is not None  # noqa: S101
    for obj_class in classes:
        if subelements := element.findall(
            f"{ns}{obj_class.get_tag_name()}",  # type: ignore[attr-defined]
        ):
            args_list.extend(
                [
                    obj_class.class_from_element(  # type: ignore[attr-defined]
                        ns=ns,
                        name_spaces=name_spaces,
                        element=subelement,
                        strict=strict,
                    )
                    for subelement in subelements
                ],
            )
    return {kwarg: args_list}



# Copyright (C) 2012 - 2023  Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
"""
Date and time handling in KML.

Any Feature in KML can have time data associated with it.
This time data has the effect of restricting the visibility of the data set to a given
time period or point in time.

https://developers.google.com/kml/documentation/time
"""
import re
from datetime import date
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import arrow

from fastkml import config
from fastkml.enums import DateTimeResolution
from fastkml.helpers import datetime_subelement
from fastkml.helpers import datetime_subelement_kwarg
from fastkml.kml_base import _BaseObject
from fastkml.registry import RegistryItem
from fastkml.registry import registry

# regular expression to parse a gYearMonth string
# year and month may be separated by an optional dash
# year is always 4 digits, month is always 2 digits
year_month_day = re.compile(
    r"^(?P<year>\d{4})(?:-)?(?P<month>\d{2})?(?:-)?(?P<day>\d{2})?$",
)


class KmlDateTime:
    """
    A KML DateTime object.

    This class is used to parse and format KML DateTime objects.

    A KML DateTime object is a string that conforms to the ISO 8601 standard
    for date and time representation. The following formats are supported:

    - yyyy-mm-ddThh:mm:sszzzzzz
    - yyyy-mm-ddThh:mm:ss
    - yyyy-mm-dd
    - yyyy-mm
    - yyyy

    The T is the separator between the date and the time, and the time zone
    is either Z (for UTC) or zzzzzz, which represents ±hh:mm in relation to
    UTC. Additionally, the value can be expressed as a date only.

    The precision of the DateTime is dictated by the DateTime value
    which can be one of the following:

    - dateTime gives second resolution
    - date gives day resolution
    - gYearMonth gives month resolution
    - gYear gives year resolution

    The KmlDateTime class can be used to parse a KML DateTime string into a
    Python datetime object, or to format a Python datetime object into a
    KML DateTime string.

    The KmlDateTime class is used by the TimeStamp and TimeSpan classes.
    """

    def __init__(
        self,
        dt: Union[date, datetime],
        resolution: Optional[DateTimeResolution] = None,
    ) -> None:
        """Initialize a KmlDateTime object."""
        if resolution is None:
            # sourcery skip: swap-if-expression
            resolution = (
                DateTimeResolution.date
                if not isinstance(dt, datetime)
                else DateTimeResolution.datetime
            )
        dt = (
            dt.date()
            if isinstance(dt, datetime) and resolution != DateTimeResolution.datetime
            else dt
        )
        if resolution == DateTimeResolution.year:
            self.dt = date(dt.year, 1, 1)
        elif resolution == DateTimeResolution.year_month:
            self.dt = date(dt.year, dt.month, 1)
        else:
            self.dt = dt
        self.resolution = (
            DateTimeResolution.date
            if not isinstance(self.dt, datetime)
            and resolution == DateTimeResolution.datetime
            else resolution
        )

    def __repr__(self) -> str:
        """Create a string (c)representation for KmlDateTime."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"dt={self.dt!r}, "
            f"resolution={self.resolution}, "
            ")"
        )

    def __bool__(self) -> bool:
        """Return True if the date or datetime is valid."""
        return isinstance(self.dt, date)

    def __eq__(self, other: object) -> bool:
        """Return True if the two objects are equal."""
        return (
            self.dt == other.dt and self.resolution == other.resolution
            if isinstance(other, KmlDateTime)
            else False
        )

    def __str__(self) -> str:
        """Return the KML DateTime string representation of the object."""
        if self.resolution == DateTimeResolution.year:
            return self.dt.strftime("%Y")
        if self.resolution == DateTimeResolution.year_month:
            return self.dt.strftime("%Y-%m")
        if self.resolution == DateTimeResolution.date:
            return (
                self.dt.date().isoformat()
                if isinstance(self.dt, datetime)
                else self.dt.isoformat()
            )
        return self.dt.isoformat()

    @classmethod
    def parse(cls, datestr: str) -> Optional["KmlDateTime"]:
        """Parse a KML DateTime string into a KmlDateTime object."""
        resolution = None
        dt = None
        if year_month_day_match := year_month_day.match(datestr):
            year = int(year_month_day_match.group("year"))
            month = int(year_month_day_match.group("month") or 1)
            day = int(year_month_day_match.group("day") or 1)
            dt = date(year, month, day)
            resolution = DateTimeResolution.date
            if year_month_day_match.group("day") is None:
                resolution = DateTimeResolution.year_month
            if year_month_day_match.group("month") is None:
                resolution = DateTimeResolution.year
        elif len(datestr) > 10:  # noqa: PLR2004
            dt = arrow.get(datestr).datetime
            resolution = DateTimeResolution.datetime
        return cls(dt, resolution) if dt else None

    @classmethod
    def get_ns_id(cls) -> str:
        """Return the namespace ID."""
        return config.KML


class _TimePrimitive(_BaseObject):
    """
    Abstract element that cannot be used directly in a KML file.

    This element is extended by the <TimeSpan> and <TimeStamp> elements.
    https://developers.google.com/kml/documentation/kmlreference#timeprimitive
    """


class TimeStamp(_TimePrimitive):
    """
    Represents a single moment in time.

    https://developers.google.com/kml/documentation/kmlreference#timestamp
    """

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        timestamp: Optional[KmlDateTime] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Times class.

        Args:
        ----
            ns : str, optional
                The namespace for the element, by default None.
            name_spaces : dict[str, str], optional
                The dictionary of namespace prefixes and URIs, by default None.
            id : str, optional
                The ID of the element, by default None.
            target_id : str, optional
                The target ID of the element, by default None.
            timestamp : KmlDateTime, optional
                The timestamp of the element, by default None.
            **kwargs : Any
                Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.timestamp = timestamp

    def __repr__(self) -> str:
        """Create a string (c)representation for TimeStamp."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"timestamp={self.timestamp!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """Return True if the timestamp is valid."""
        return bool(self.timestamp)


registry.register(
    TimeStamp,
    item=RegistryItem(
        ns_ids=("kml", "gx", ""),
        classes=(KmlDateTime,),
        attr_name="timestamp",
        node_name="when",
        get_kwarg=datetime_subelement_kwarg,
        set_element=datetime_subelement,
    ),
)


class TimeSpan(_TimePrimitive):
    """
    Represents an extent in time bounded by begin and end dateTimes.

    https://developers.google.com/kml/documentation/kmlreference#timespan
    """

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        begin: Optional[KmlDateTime] = None,
        end: Optional[KmlDateTime] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a new instance of the Times class.

        Args:
        ----
            ns (Optional[str]): The namespace for the element.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the element.
            target_id (Optional[str]): The target ID of the element.
            begin (Optional[KmlDateTime]): The begin time.
            end (Optional[KmlDateTime]): The end time.
            **kwargs (Any): Additional keyword arguments.

        Returns:
        -------
            None

        """
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            **kwargs,
        )
        self.begin = begin
        self.end = end

    def __repr__(self) -> str:
        """Create a string (c)representation for TimeSpan."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"begin={self.begin!r}, "
            f"end={self.end!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __bool__(self) -> bool:
        """Return True if the begin or end date is valid."""
        return bool(self.begin) or bool(self.end)


registry.register(
    TimeSpan,
    item=RegistryItem(
        ns_ids=("kml", "gx", ""),
        classes=(KmlDateTime,),
        attr_name="begin",
        node_name="begin",
        get_kwarg=datetime_subelement_kwarg,
        set_element=datetime_subelement,
    ),
)
registry.register(
    TimeSpan,
    item=RegistryItem(
        ns_ids=("kml", "gx", ""),
        classes=(KmlDateTime,),
        attr_name="end",
        node_name="end",
        get_kwarg=datetime_subelement_kwarg,
        set_element=datetime_subelement,
    ),
)



# Copyright (C) 2012 - 2024 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""Abstract base classes."""
import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import cast

from typing_extensions import Self

from fastkml import config
from fastkml.enums import Verbosity
from fastkml.registry import registry
from fastkml.types import Element

logger = logging.getLogger(__name__)

__all__ = ["_XMLObject"]


class _XMLObject:
    """XML Baseclass."""

    _default_nsid: str = ""
    _node_name: str = ""
    name_spaces: Dict[str, str]
    __kwarg_keys: Tuple[str, ...]

    def __init__(
        self,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the XML Object.

        Parameters
        ----------
        ns : Optional[str], default=None
            The namespace of the XML object.
        name_spaces : Optional[Dict[str, str]], default=None
            The dictionary of namespace prefixes and URIs.
        **kwargs : Any
            Additional keyword arguments.

        """
        name_spaces = name_spaces or {}
        self.name_spaces = {**config.NAME_SPACES, **name_spaces}
        self.ns: str = (
            self.name_spaces.get(self._default_nsid, "") if ns is None else ns
        )
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])
        self.__kwarg_keys = tuple(kwargs.keys())

    def __repr__(self) -> str:
        """
        Create a string (c)representation for _XMLObject.

        Returns
        -------
        str
            The string representation of the object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    def __str__(self) -> str:
        """
        Return the string representation of the object.

        Returns
        -------
        str
            The string representation of the object.

        """
        return self.to_string()

    def __eq__(self, other: object) -> bool:
        """
        Compare two _XMLObject instances for equality.

        Parameters
        ----------
        other : object
            The object to compare with.

        Returns
        -------
        bool
            True if the objects are equal, False otherwise.

        """
        return self.__dict__ == other.__dict__ if type(self) is type(other) else False

    def etree_element(
        self,
        precision: Optional[int] = None,
        verbosity: Verbosity = Verbosity.normal,
    ) -> Element:
        """
        Return the KML Object as an Element.

        Parameters
        ----------
        precision : Optional[int], default=None
            The precision of the KML object.
        verbosity : Verbosity, default=Verbosity.normal
            The verbosity level.

        Returns
        -------
        Element
            The KML object as an Element.

        """
        element: Element = config.etree.Element(
            f"{self.ns}{self.get_tag_name()}",
        )
        for item in registry.get(self.__class__):
            item.set_element(
                obj=self,
                element=element,
                attr_name=item.attr_name,
                node_name=item.node_name,
                precision=precision,
                verbosity=verbosity,
                default=item.default,
            )
        return element

    def to_string(
        self,
        *,
        prettyprint: bool = True,
        precision: Optional[int] = None,
        verbosity: Verbosity = Verbosity.normal,
    ) -> str:
        """
        Return the KML Object as serialized xml.

        Parameters
        ----------
        prettyprint : bool, default=True
            Whether to pretty print the XML.
        precision : Optional[int], default=None
            The precision of the KML object.
        verbosity : Verbosity, default=Verbosity.normal
            The verbosity level.

        Returns
        -------
        str
            The KML object as serialized XML.

        """
        element = self.etree_element(
            precision=precision,
            verbosity=verbosity,
        )
        try:
            return cast(
                str,
                config.etree.tostring(
                    element,
                    encoding="UTF-8",
                    pretty_print=prettyprint,
                ).decode(
                    "UTF-8",
                ),
            )
        except TypeError:
            return cast(
                str,
                config.etree.tostring(
                    element,
                    encoding="UTF-8",
                ).decode(
                    "UTF-8",
                ),
            )

    def _get_splat(self) -> Dict[str, Any]:
        """
        Get the keyword arguments as a dictionary.

        Returns
        -------
        Dict[str, Any]
            The keyword arguments as a dictionary.

        """
        return {
            key: getattr(self, key)
            for key in self.__kwarg_keys
            if getattr(self, key) is not None
        }

    @classmethod
    def get_tag_name(cls) -> str:
        """
        Return the tag name.

        Returns
        -------
        str
            The tag name.

        """
        return cls.__name__

    @classmethod
    def _get_ns(cls, ns: Optional[str], name_spaces: Dict[str, str]) -> str:
        """
        Get the namespace.

        Parameters
        ----------
        ns : Optional[str]
            The namespace.
        name_spaces : Dict[str, str]
            The dictionary of namespace prefixes and URIs.

        Returns
        -------
        str
            The namespace.

        """
        return name_spaces.get(cls._default_nsid, "") if ns is None else ns

    @classmethod
    def _get_kwargs(
        cls,
        *,
        ns: str,
        name_spaces: Optional[Dict[str, str]] = None,
        element: Element,
        strict: bool,
    ) -> Dict[str, Any]:
        """
        Get the keyword arguments for the class constructor.

        Parameters
        ----------
        ns : str
            The namespace.
        name_spaces : Optional[Dict[str, str]], default=None
            The dictionary of namespace prefixes and URIs.
        element : Element
            The XML element.
        strict : bool
            Whether to enforce strict parsing.

        Returns
        -------
        Dict[str, Any]
            The keyword arguments for the class constructor.

        """
        name_spaces = name_spaces or {}
        name_spaces = {**config.NAME_SPACES, **name_spaces}
        kwargs: Dict[str, Any] = {"ns": ns, "name_spaces": name_spaces}
        for item in registry.get(cls):
            for name_space in item.ns_ids:
                kwarg = item.get_kwarg(
                    element=element,
                    ns=name_spaces.get(name_space, ""),
                    name_spaces=name_spaces,
                    node_name=item.node_name,
                    kwarg=item.attr_name,
                    classes=item.classes,
                    strict=strict,
                )
                if kwarg:
                    kwargs.update(
                        kwarg,
                    )
                    break
        return kwargs

    @classmethod
    def class_from_element(
        cls,
        *,
        ns: str,
        name_spaces: Optional[Dict[str, str]] = None,
        element: Element,
        strict: bool,
    ) -> Self:
        """
        Create an XML object from an etree element.

        Parameters
        ----------
        ns : str
            The namespace.
        name_spaces : Optional[Dict[str, str]], default=None
            The dictionary of namespace prefixes and URIs.
        element : Element
            The XML element.
        strict : bool
            Whether to enforce strict parsing.

        Returns
        -------
        Self
            The XML object.

        """
        kwargs = cls._get_kwargs(
            ns=ns,
            name_spaces=name_spaces,
            element=element,
            strict=strict,
        )
        return cls(
            **kwargs,
        )

    @classmethod
    def from_string(
        cls,
        string: str,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        strict: bool = True,
    ) -> Self:
        """
        Create an XML object from a string.

        Parameters
        ----------
        string : str
            The string representation of the XML object.
        ns : Optional[str], default=None
            The namespace of the XML object.
        name_spaces : Optional[Dict[str, str]], default=None
            The dictionary of namespace prefixes and URIs.
        strict : bool, default=True
            Whether to enforce strict parsing.

        Returns
        -------
        Self
            The XML object.

        """
        name_spaces = name_spaces or {}
        name_spaces = {**config.NAME_SPACES, **name_spaces}
        ns = cls._get_ns(ns, name_spaces=name_spaces)
        return cls.class_from_element(
            ns=ns,
            name_spaces=name_spaces,
            strict=strict,
            element=cast(
                Element,
                config.etree.fromstring(string),
            ),
        )



# Copyright (C) 2012 - 2023 Christian Ledermann
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA

"""
KML Extension Namespace and the gx prefix.

With the launch of Google Earth 5.0, Google has provided extensions to KML
to support a number of new features. These extensions use the gx prefix
and the following namespace URI::

    xmlns:gx="http://www.google.com/kml/ext/2.2"

This namespace URI must be added to the <kml> element in any KML file
using gx-prefixed elements::

    <kml
        xmlns="http://www.opengis.net/kml/2.2"
        xmlns:gx="http://www.google.com/kml/ext/2.2"
    >

Extensions to KML may not be supported in all geo-browsers. If your
browser doesn't support particular extensions, the data in those
extensions should be silently ignored, and the rest of the KML file
should load without errors.

Elements that currently use the gx prefix are:

* gx:altitudeMode
* gx:altitudeOffset
* gx:angles
* gx:AnimatedUpdate
* gx:balloonVisibility
* gx:coord
* gx:delayedStart
* gx:drawOrder
* gx:duration
* gx:FlyTo
* gx:flyToMode
* gx:h
* gx:horizFov
* gx:interpolate
* gx:labelVisibility
* gx:LatLonQuad
* gx:MultiTrack
* gx:vieweroptions
* gx:outerColor
* gx:outerWidth
* gx:physicalWidth
* gx:Playlist
* gx:playMode
* gx:SoundCue
* gx:TimeSpan
* gx:TimeStamp
* gx:Tour
* gx:TourControl
* gx:TourPrimitive
* gx:Track
* gx:ViewerOptions
* gx:w
* gx:Wait
* gx:x
* gx:y

The complete XML schema for elements in this extension namespace is
located at http://developers.google.com/kml/schema/kml22gx.xsd.
"""
import logging
from dataclasses import dataclass
from itertools import zip_longest
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import pygeoif.geometry as geo
from pygeoif.types import PointType

from fastkml import config
from fastkml.enums import AltitudeMode
from fastkml.geometry import _Geometry
from fastkml.helpers import bool_subelement
from fastkml.helpers import coords_subelement_list
from fastkml.helpers import coords_subelement_list_kwarg
from fastkml.helpers import datetime_subelement_list
from fastkml.helpers import datetime_subelement_list_kwarg
from fastkml.helpers import enum_subelement
from fastkml.helpers import subelement_bool_kwarg
from fastkml.helpers import subelement_enum_kwarg
from fastkml.helpers import xml_subelement_list
from fastkml.helpers import xml_subelement_list_kwarg
from fastkml.registry import RegistryItem
from fastkml.registry import registry
from fastkml.times import KmlDateTime

__all__ = [
    "Angle",
    "MultiTrack",
    "Track",
    "TrackItem",
    "track_items_to_geometry",
    "tracks_to_geometry",
]

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Angle:
    """
    The gx:angles element specifies the heading, tilt, and roll.

    The angles are specified in degrees, and the
    default values are 0 (heading and tilt) and 0 (roll). The angles
    are specified in the following order: heading, tilt, roll.
    """

    heading: float = 0.0
    tilt: float = 0.0
    roll: float = 0.0

    @property
    def coords(self) -> PointType:
        """
        Get the coordinates of the angle.

        Returns
        -------
        PointType
            The coordinates of the angle.

        """
        return (self.heading, self.tilt, self.roll)


@dataclass(frozen=True)
class TrackItem:
    """A track item describes an objects position and heading at a specific time."""

    when: KmlDateTime
    coord: geo.Point
    angle: Optional[Angle] = None


def track_items_to_geometry(track_items: Iterable[TrackItem]) -> geo.LineString:
    """
    Convert a sequence of TrackItems to a LineString geometry.

    Args:
    ----
        track_items : Iterable[TrackItem]
            A sequence of TrackItems.

    Returns:
    -------
        geo.LineString
            A LineString geometry representing the track.

    """
    return geo.LineString.from_points(
        *[item.coord for item in track_items if item.coord is not None],
    )


class Track(_Geometry):
    """
    A track describes how an object moves through the world over a given time period.

    This feature allows you to create one visible object in Google Earth
    (either a Point icon or a Model) that encodes multiple positions for the same object
    for multiple times. In Google Earth, the time slider allows the user to move the
    view through time, which animates the position of the object.

    Tracks are a more efficient mechanism for associating time data with visible
    Features, since you create only one Feature, which can be associated with multiple
    time elements as the object moves through space.

    https://developers.google.com/kml/documentation/kmlreference#gxtrack
    """

    _default_nsid = config.GX
    track_items: List[TrackItem]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        track_items: Optional[Iterable[TrackItem]] = None,
        whens: Optional[Iterable[KmlDateTime]] = None,
        coords: Optional[Iterable[PointType]] = None,
        angles: Optional[Iterable[PointType]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a GX object.

        Parameters
        ----------
        ns : Optional[str], optional
            The namespace for the GX object, by default None
        name_spaces : Optional[Dict[str, str]], optional
            A dictionary of namespace prefixes and URIs, by default None
        id : Optional[str], optional
            The ID of the GX object, by default None
        target_id : Optional[str], optional
            The target ID of the GX object, by default None
        altitude_mode : Optional[AltitudeMode], optional
            The altitude mode of the GX object, by default None
        track_items : Optional[Iterable[TrackItem]], optional
            The track items of the GX object, by default None
        whens : Optional[Iterable[KmlDateTime]], optional
            The timestamps of the track items, by default None
        coords : Optional[Iterable[PointType]], optional
            The coordinates of the track items, by default None
        angles : Optional[Iterable[PointType]], optional
            The angles of the track items, by default None
        **kwargs : Any, optional
            Additional keyword arguments.

        Raises
        ------
        ValueError
            If both `geometry` and `track_items` are specified.

        """
        angles = list(angles) if angles else []
        if (whens or coords) and track_items:
            msg = "Cannot specify both geometry and track_items"
            raise ValueError(msg)
        if not track_items and whens and coords:
            track_items = [
                TrackItem(
                    when=cast(KmlDateTime, when),
                    coord=geo.Point(*coord),
                    angle=Angle(*angle),
                )
                for when, coord, angle in zip_longest(
                    whens,
                    coords,
                    angles,
                    fillvalue=(),
                )
            ]
        self.track_items = list(track_items) if track_items else []
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            altitude_mode=altitude_mode,
            **kwargs,
        )

    def __repr__(self) -> str:
        """
        Create a string representation for Track.

        Returns
        -------
        str
            The string representation of the Track object.

        """
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"track_items={self.track_items!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    @property
    def geometry(self) -> Optional[geo.LineString]:
        """
        Get the geometry of the track.

        Returns
        -------
        Optional[geo.LineString]
            The geometry of the track.

        """
        return track_items_to_geometry(self.track_items)

    @property
    def whens(self) -> Tuple[KmlDateTime, ...]:
        """
        Get the timestamps of the track items.

        Returns
        -------
        Tuple[KmlDateTime]
            The timestamps of the track items.

        """
        return tuple(item.when for item in self.track_items)

    @property
    def coords(self) -> Tuple[PointType, ...]:
        """
        Get the coordinates of the track items.

        Returns
        -------
        Tuple[PointType]
            The coordinates of the track items.

        """
        return tuple(
            item.coord.coords[0]  # type: ignore[misc]
            for item in self.track_items
            if item.coord
        )

    @property
    def angles(self) -> Tuple[PointType, ...]:
        """
        Get the angles of the track items.

        Returns
        -------
        Tuple[Angle]
            The angles of the track items.

        """
        return tuple(item.angle.coords for item in self.track_items if item.angle)

    def __bool__(self) -> bool:
        """
        Check if the track has any track items.

        Returns
        -------
        bool
            True if the track has track items, False otherwise.

        """
        return bool(self.track_items)


registry.register(
    Track,
    item=RegistryItem(
        ns_ids=("gx", "kml", ""),
        classes=(AltitudeMode,),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)
registry.register(
    Track,
    item=RegistryItem(
        ns_ids=("kml", "gx", ""),
        classes=(KmlDateTime,),
        attr_name="whens",
        node_name="when",
        get_kwarg=datetime_subelement_list_kwarg,
        set_element=datetime_subelement_list,
    ),
)
registry.register(
    Track,
    item=RegistryItem(
        ns_ids=("gx", ""),
        classes=(tuple,),
        attr_name="coords",
        node_name="coord",
        get_kwarg=coords_subelement_list_kwarg,
        set_element=coords_subelement_list,
    ),
)
registry.register(
    Track,
    item=RegistryItem(
        ns_ids=("gx", ""),
        classes=(tuple,),
        attr_name="angles",
        node_name="angles",
        get_kwarg=coords_subelement_list_kwarg,
        set_element=coords_subelement_list,
        default=(0.0, 0.0, 0.0),
    ),
)


def tracks_to_geometry(tracks: Iterable[Track]) -> geo.MultiLineString:
    """
    Convert a collection of tracks to a MultiLineString geometry.

    Args:
    ----
        tracks : Iterable[Track]
            A collection of tracks.

    Returns:
    -------
        geo.MultiLineString
            A MultiLineString geometry representing the tracks.

    """
    return geo.MultiLineString.from_linestrings(
        *[track.geometry for track in tracks if track.geometry],
    )


class MultiTrack(_Geometry):
    """
    A MultiTrack is a collection of tracks.

    A multi-track element is used to combine multiple track elements into a single
    conceptual unit. For example, suppose you collect GPS data for a day's bike ride
    that includes several rest stops and a stop for lunch. Because of the interruptions
    in time, one bike ride might appear as four different tracks when the times and
    positions are plotted.
    Grouping these <gx:Track> elements into one <gx:MultiTrack> container causes them to
    be displayed in Google Earth as sections of a single path.
    When the icon reaches the end of one segment, it moves to the beginning of the next
    segment.
    The <gx:interpolate> element specifies whether to stop at the end of one track and
    jump immediately to the start of the next one, or to interpolate the missing values
    between the two tracks.
    """

    _default_nsid = config.GX
    tracks: List[Track]

    def __init__(
        self,
        *,
        ns: Optional[str] = None,
        name_spaces: Optional[Dict[str, str]] = None,
        id: Optional[str] = None,
        target_id: Optional[str] = None,
        altitude_mode: Optional[AltitudeMode] = None,
        tracks: Optional[Iterable[Track]] = None,
        interpolate: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize a GX object.

        Args:
        ----
            ns (Optional[str]): The namespace for the GX object.
            name_spaces (Optional[Dict[str, str]]): The dictionary of namespace prefixes
                and URIs.
            id (Optional[str]): The ID of the GX object.
            target_id (Optional[str]): The target ID of the GX object.
            altitude_mode (Optional[AltitudeMode]): The altitude mode of the GX object.
            geometry (Optional[geo.MultiLineString]): The geometry of the GX object.
            tracks (Optional[Iterable[Track]]): The tracks of the GX object.
            interpolate (Optional[bool]): The interpolate flag of the GX object.
            **kwargs (Any): Additional keyword arguments.

        Raises:
        ------
            ValueError: If both geometry and tracks are specified.

        """
        self.tracks = [t for t in tracks if t] if tracks else []
        self.interpolate = interpolate
        super().__init__(
            ns=ns,
            name_spaces=name_spaces,
            id=id,
            target_id=target_id,
            altitude_mode=altitude_mode,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Create a string (c)representation for MultiTrack."""
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}("
            f"ns={self.ns!r}, "
            f"name_spaces={self.name_spaces!r}, "
            f"id={self.id!r}, "
            f"target_id={self.target_id!r}, "
            f"altitude_mode={self.altitude_mode}, "
            f"tracks={self.tracks!r}, "
            f"interpolate={self.interpolate!r}, "
            f"**{self._get_splat()!r},"
            ")"
        )

    @property
    def geometry(self) -> Optional[geo.MultiLineString]:
        """
        Get the geometry of the gx object.

        Returns
        -------
            Optional[geo.MultiLineString]: The geometry of the gx object.

        """
        return tracks_to_geometry(self.tracks)

    def __bool__(self) -> bool:
        """
        Check if the object has any tracks.

        Returns
        -------
        bool
            True if the object has tracks, False otherwise.

        """
        return bool(self.tracks)


registry.register(
    MultiTrack,
    item=RegistryItem(
        ns_ids=("gx", "kml"),
        classes=(AltitudeMode,),
        attr_name="altitude_mode",
        node_name="altitudeMode",
        get_kwarg=subelement_enum_kwarg,
        set_element=enum_subelement,
        default=AltitudeMode.clamp_to_ground,
    ),
)
registry.register(
    MultiTrack,
    item=RegistryItem(
        ns_ids=("gx",),
        classes=(bool,),
        attr_name="interpolate",
        node_name="interpolate",
        get_kwarg=subelement_bool_kwarg,
        set_element=bool_subelement,
    ),
)
registry.register(
    MultiTrack,
    item=RegistryItem(
        ns_ids=("gx",),
        classes=(Track,),
        attr_name="tracks",
        node_name="gx:Track",
        get_kwarg=xml_subelement_list_kwarg,
        set_element=xml_subelement_list,
    ),
)
